import React, {startTransition, useCallback, useContext, useEffect, useMemo, useRef, useState} from "react";
import ReactDOM from "react-dom";
import {v4 as uuidv4} from "uuid";

/**
 * Notification UUID type.
 */
export type NotificationUuid = string;

/**
 * Notification data.
 */
export interface NotificationData
{
	/**
	 * The notification content.
	 */
	content: React.ReactNode;
}

/**
 * Type of notification emitter function.
 */
export type EmitNotificationFunction = (content: React.ReactNode) => void;

/**
 * Type of notification close function.
 */
export type RemoveNotificationFunction = (uuid: NotificationUuid) => void;

/**
 * Type of notification close function.
 */
export type CloseNotificationFunction = (uuid: NotificationUuid) => void;

/**
 * Type of notification closed state function.
 */
export type IsNotificationClosedFunction = (uuid: NotificationUuid) => boolean;


/**
 * Interface of notifications state.
 */
export interface NotificationsContextState
{
	/**
	 * Notification function.
	 * @param content Notification content.
	 */
	notify: EmitNotificationFunction;

	/**
	 * Close notification function.
	 * @param uuid UUID of notification to close.
	 */
	close: CloseNotificationFunction;

	/**
	 * Is given notification closed?
	 * @param uuid UUID of notification to get closed state.
	 */
	isClosed: IsNotificationClosedFunction;
}

export const NotificationsContext = React.createContext<NotificationsContextState>({
	notify() {},
	close() {},
	isClosed() { return false; },
});

/**
 * Hook to emit a notification.
 */
export function useNotify(): EmitNotificationFunction
{
	return useContext(NotificationsContext).notify;
}



/**
 * Notifications provider.
 */
export function NotificationsProvider({children}: React.PropsWithChildren<{}>)
{
	// Notifications.
	const [notifications, setNotifications] = useState<Record<NotificationUuid, NotificationData>>({});
	// Keeping track of closed notifications that are still on (while transitioning out).
	const [closedNotifications, setClosedNotifications] = useState<Record<NotificationUuid, boolean>>()

	// Initialize remove notification function.
	const remove = useRef<RemoveNotificationFunction>();
	remove.current = useCallback((uuid) => {
		// Copy the notifications list.
		const newNotifications = {...notifications};
		const newClosedNotifications = {...closedNotifications};
		// Remove the given notification from the list.
		delete newNotifications[uuid];
		delete newClosedNotifications[uuid];
		// Set the new notifications list.
		setNotifications(newNotifications);
		setClosedNotifications(newClosedNotifications);
	}, [notifications, setNotifications, closedNotifications, setClosedNotifications]);

	// Initialize close notification function with animation.
	const close = useRef<CloseNotificationFunction>();
	close.current = useCallback((uuid) => {
		// Add the given curtain UUID to the list of closed curtains.
		setClosedNotifications({
			...closedNotifications,
			[uuid]: true,
		});

		// Remove the notification 300ms later.
		window.setTimeout(() => {
			// Remove the curtain.
			remove.current(uuid);
		}, 300);
	}, [remove, closedNotifications, setClosedNotifications]);

	// Initialize isClosed notification function.
	const isClosed = useRef<IsNotificationClosedFunction>();
	isClosed.current = useCallback((uuid) => (!!closedNotifications?.[uuid]), [closedNotifications]);

	// Initialize notify function.
	const notify = useRef<EmitNotificationFunction>();
	notify.current = useCallback((content: React.ReactNode) => {
		// Generate a new notification UUID for the new notification.
		const notificationUuid = uuidv4();

		// Add the notification to the list of notifications, with the generated UUID.
		setNotifications({
			...notifications,
			[notificationUuid]: {
				content: content,
			},
		});

		// Close notification 10s later.
		setTimeout(() => {
			close.current(notificationUuid);
		}, 10000);
	}, [notifications, setNotifications]);

	// Initialize context state from action functions.
	const contextState = useMemo(() => ({
		notify: (content: React.ReactNode) => notify.current(content),
		close: (uuid: NotificationUuid) => close.current(uuid),
		isClosed: (uuid: NotificationUuid) => isClosed.current(uuid),
	}), [notify, isClosed]);

	return (
		<NotificationsContext.Provider value={contextState}>
			{children}
			<NotificationsPortal notifications={notifications} />
		</NotificationsContext.Provider>
	);
}

/**
 * Curtains portal manager.
 */
function NotificationsPortal({notifications}: {
	notifications: Record<NotificationUuid, NotificationData>;
})
{
	return ReactDOM.createPortal((
		<ul className={"notifications"}>
			{ // Show notifications list.
				Object.entries(notifications).map(([uuid, notificationData]) => (
					<NotificationInstance key={uuid} uuid={uuid}>
						{notificationData.content}
					</NotificationInstance>
				))
			}
		</ul>
	), document.body);
}

/**
 * A notification context.
 */
export const NotificationContext = React.createContext<{
	/**
	 * Notification UUID.
	 */
	uuid: NotificationUuid;

	/**
	 * Notification closed state.
	 */
	closed: boolean;
}>(undefined);

/**
 * Notification component.
 */
function NotificationInstance({uuid, children}: React.PropsWithChildren<{
	/**
	 * Notification UUID.
	 */
	uuid: NotificationUuid;
}>)
{
	// Get notifications context.
	const {isClosed} = useContext(NotificationsContext);

	return (
		<NotificationContext.Provider value={{
			uuid: uuid,
			closed: isClosed(uuid),
		}}>
			{children}
		</NotificationContext.Provider>
	);
}
